/*
 * Modified from board-htcwizard.c
 *
 * HTC Prophet init stuff
 * Copyright (C) 2006 Unai Uribarri
 * Copyright (C) 2008 linwizard.sourceforge.net
 * Copyright (C) 2009 Peter Zotov <whitequark@whitequark.ru>
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify  it under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 * 
 * You should have  received a copy of the  GNU General Public License
 * along  with  this program;  if  not,  write  to the  Free  Software
 * Foundation,  Inc.,  51 Franklin  Street,  Fifth  Floor, Boston,  MA
 * 02110-1301, USA. 
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/bootmem.h>
#include <linux/i2c-gpio.h>

#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/arch/omap850.h>
#include <asm/page.h>
#include <asm/memory.h>
#include <asm/arch/common.h>
#include <asm/arch/board.h>

#ifdef CONFIG_EFB_DEBUG
#include <asm/arch/efb.h>
#endif

#include <asm/arch/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch/gpio.h>
#include <asm/arch/keypad.h>
#include <asm/arch/spi100k.h>

#include <linux/spi/spi.h>
#include <linux/spi/tsc2046.h>

#include <linux/delay.h>

#ifdef CONFIG_HTC_MMC
#include "htc-mmc.h"
#endif

#define HTCPROPHET_GPIO_DM 35
#define HTCPROPHET_GPIO_DP 36
#define HTCPROPHET_GPIO_POWER 139

static struct omap_lcd_config htcprophet_lcd_config __initdata = {
	.ctrl_name	= "internal",
};

static struct omap_usb_config htcprophet_usb_config __initdata = {
	.otg		= 0,
	.register_host	= 0,
	.register_dev	= 1,
	.hmc_mode	= 4,
	.pins[0]	= 2,
};

static struct omap_mmc_config htcprophet_mmc_config __initdata =
{
	.mmc[0] = {
		.enabled = 1,
		.nomux = 1,
		.wire4 = 1,
		.power_pin = -1,
		.switch_pin = -1,
	}
};

/*
  UART1 Bluetooth
  UART2 IR?
*/
static struct omap_uart_config htcprophet_uart_config __initdata = {
        .enabled_uarts = (1 << 0) | (1 << 1),
};

static struct omap_board_config_kernel htcprophet_config[] __initdata = 
{
	{ OMAP_TAG_LCD, &htcprophet_lcd_config },
	{ OMAP_TAG_USB, &htcprophet_usb_config },
	{ OMAP_TAG_MMC, &htcprophet_mmc_config },
	{ OMAP_TAG_UART, &htcprophet_uart_config },
};

/* GSM device */
#define PROPHET_GSM_PHYS_START          0x13c00000 /* end of RAM */
#define PROPHET_GSM_PHYS_SIZE           0x00400000 /* 4.0 MB code/data/fifo */

static struct resource prophet_gsm_resources[] = {
	{       /* GSM DSP MMU */
                .start          = IO_ADDRESS(OMAP850_DSP_MMU_BASE),
                .end            = IO_ADDRESS(OMAP850_DSP_MMU_BASE) + 0x54,
                .flags          = IORESOURCE_MEM,
        },
        {       /* GSM software interrupt */
                .start          = INT_850_ICR,
                .flags          = IORESOURCE_IRQ,
        },
        {       /* GSM radio interrupt */
                .start          = INT_850_DBB_RF_EN,
                .flags          = IORESOURCE_IRQ,
        },
};

static struct platform_device gsm_device = {
        .name           = "wizard-gsm",
        .id             = 1,
        .num_resources  = ARRAY_SIZE(prophet_gsm_resources),
        .resource       = prophet_gsm_resources,
};

/* Keyboard definition */
static int htc_prophet_keymap[] = {
   KEY(0, 0, KEY_LEFTCTRL), // communication menu in windows - search
   KEY(0, 1, KEY_CAMERA), // camera button
   KEY(0, 3, KEY_VOLUMEUP), // volume +
   KEY(6, 3, KEY_VOLUMEDOWN), // volume -
   KEY(0, 6, KEY_F1), // first screen key - back
   KEY(6, 6, KEY_F2), // second screen key - menu
   KEY(0, 2, KEY_PHONE), // call button
   KEY(6, 2, KEY_CANCEL), // hangup button
   KEY(0, 5, KEY_MAIL), // windows button - home
   KEY(6, 5, KEY_SLASH), // ok button
   0
};

struct omap_kp_platform_data htcprophet_kp_data = {
	.rows	= 7,
	.cols	= 7,
	.delay = 20,
	.rep = 1,
	.keymap = htc_prophet_keymap,
};

static struct resource kp_resources[] = {
	[0] = {
		.start	= INT_850_MPUIO_KEYPAD,
		.end	= INT_850_MPUIO_KEYPAD,
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device kp_device = {
	.name		= "omap-keypad",
	.id		= -1,
	.dev		= {
		.platform_data = &htcprophet_kp_data,
	},
	.num_resources	= ARRAY_SIZE(kp_resources),
	.resource	= kp_resources,
};

/* LCD Device resources */
static struct platform_device lcd_device = {
	.name		= "lcd_htcprophet",
	.id		= -1,
};

#if defined(CONFIG_I2C_GPIO) || defined(CONFIG_I2C_GPIO_MODULE)

static struct i2c_gpio_platform_data i2cgpio_device_data = {
	.sda_pin		= 69,
	.scl_pin		= 70,
	.udelay			= 2,		/* ~100 kHz */
};

static struct platform_device i2cgpio_device = {
	.name			= "i2c-gpio",
	.id			= 0,
	.dev	 = {
		.platform_data	= &i2cgpio_device_data,
	},
};

#endif

struct platform_device htcprophet_bt = {
        .name = "htc-bt",
        .id = 1
};

#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)
#include <linux/input.h>
#include <linux/gpio_keys.h>

static struct gpio_keys_button prophet_gpio_keys_table[] = {
   { KEY_POWER, HTCPROPHET_GPIO_POWER, 1, "power button", EV_KEY, 1 },
};

static struct gpio_keys_platform_data prophet_gpio_keys_data = {
   .buttons    = prophet_gpio_keys_table,
   .nbuttons   = ARRAY_SIZE(prophet_gpio_keys_table),
};

static struct platform_device prophet_gpio_keys_device = {
   .name       = "gpio-keys",
   .dev        = {
       .platform_data = &prophet_gpio_keys_data,
   },
};
#endif

static struct platform_device *devices[] __initdata = {
 	&gsm_device,
	&kp_device,
	&lcd_device,
#if defined(CONFIG_HTC_BT)
	&htcprophet_bt,
#endif
#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)
   &prophet_gpio_keys_device,
#endif
};

/*
 * Touchscreen
 */
static struct tsc2046_platform_data htcprophet_ts_platform_data __initdata = {
        .ts_x_plate_ohm    = 496, 
        .dav_gpio          = 76,
        .gpio_debounce     = 0,
		  .ts_max_pressure   = 100000,
        .ts_touch_pressure = 5000, 
};

static struct spi_board_info htcprophet_spi_board_info[] __initdata = { 
	{
 	       .modalias               = "tsc2046",
 	       .platform_data          = &htcprophet_ts_platform_data,
	       .max_speed_hz           = 2500000,
	       .bus_num                = 2,
	       .chip_select            = 1,
	       .irq		       = OMAP_GPIO_IRQ(76),
	} 
};
/* LCD register definition (maybe to move somewhere else */
#define       OMAP_LCDC_CONTROL               (0xfffec000 + 0x00)
#define       OMAP_LCDC_STATUS                (0xfffec000 + 0x10)
#define       OMAP_DMA_LCD_CCR                (0xfffee300 + 0xc2)
#define       OMAP_DMA_LCD_CTRL               (0xfffee300 + 0xc4)
#define       OMAP_LCDC_CTRL_LCD_EN           (1 << 0)
#define       OMAP_LCDC_STAT_DONE             (1 << 0)
static void htcprophet_lcd_init(void)
{
        u32 reg;
        /* disable controller if active */
        reg = omap_readl(OMAP_LCDC_CONTROL);
        if (reg & OMAP_LCDC_CTRL_LCD_EN) {
                reg &= ~OMAP_LCDC_CTRL_LCD_EN;
                omap_writel(reg, OMAP_LCDC_CONTROL);

                /* wait for end of frame */
                while (!(omap_readl(OMAP_LCDC_STATUS) & OMAP_LCDC_STAT_DONE));

                /* turn off DMA */
                reg = omap_readw(OMAP_DMA_LCD_CCR);
                reg &= ~(1 << 7);
                omap_writew(reg, OMAP_DMA_LCD_CCR);

                reg = omap_readw(OMAP_DMA_LCD_CTRL);
                reg &= ~(1 << 8);
                omap_writew(reg, OMAP_DMA_LCD_CTRL);
        }
}


/*
 * Init functions from here on
 */
static void __init htcprophet_map_io(void)
{
	omap1_map_common_io();
	printk("htcprophet_map_io done.\n");

#ifdef CONFIG_EFB_DEBUG
	efb_enable();
#endif
	htcprophet_lcd_init();

	/* Reserve GSM Memory */
	reserve_bootmem(PROPHET_GSM_PHYS_START, PROPHET_GSM_PHYS_SIZE, 0);
}

static void __init htcprophet_disable_watchdog(void)
{
  /* Disable watchdog if running */
  if (omap_readl(OMAP_WDT_TIMER_MODE) & 0x8000) {
    /*
     * disable a potentially running watchdog timer before
     * it kills us.
     */
    printk("OMAP850 Watchdog seems to be activated, disabling it for now.\n");
    omap_writel(0xF5, OMAP_WDT_TIMER_MODE);
    omap_writel(0xA0, OMAP_WDT_TIMER_MODE);
  }
}

/* TSC2046 init from board-nokia770.c */


static void __init htcprophet_usb_enable(void)
{
	unsigned int tries = 20;
	printk("Trying to enable USB.\n");

	/* clock configuration */
	omap_writew(omap_readw(ULPD_SOFT_REQ) | (1 << 8) | SOFT_USB_CLK_REQ, ULPD_SOFT_REQ);

	/*  clk_enable(&l3_ocpi_ck); */
	omap_writew(omap_readw(ARM_IDLECT3) | (1 << 0), ARM_IDLECT3);

	/* setup USB_VBUS_CTRL, USB_VBUS_MODE, USB_TRANSCEIVER_SPEED, USB_PWRDN, USB_TRANSCEIVER_SEL */
	omap_writel(omap_readl(OMAP730_MODE_1) & 0xfefff3ff, OMAP730_MODE_1);
	omap_writel(omap_readl(OMAP730_MODE_1) | 0x06808000, OMAP730_MODE_1);

	/* GPIO 31 -> USB_DM */
	/* GPIO 32 -> USB_DP */
	/* GPIO 33 -> USB_PU_EN */
	omap_writel(omap_readl(OMAP730_IO_CONF_2) & 0xf11fffff, OMAP730_IO_CONF_2);
	omap_writel(omap_readl(OMAP730_IO_CONF_2) | 0x01100000, OMAP730_IO_CONF_2);

}

static void __init htcprophet_spi_mux(void)
{
	/* Setup MUX config for SPI */
	omap_writel(omap_readl(OMAP850_IO_CONF_6) |  0x00088880, OMAP850_IO_CONF_6);
	omap_writel(omap_readl(OMAP850_IO_CONF_6) & ~0x00077770, OMAP850_IO_CONF_6);

	omap_writel(omap_readl(OMAP850_IO_CONF_8) |  0x01000000, OMAP850_IO_CONF_8);
	omap_writel(omap_readl(OMAP850_IO_CONF_8) & ~0x10110000, OMAP850_IO_CONF_8);

	omap_writel(omap_readl(OMAP850_IO_CONF_9) |  0x00000010, OMAP850_IO_CONF_9);
	omap_writel(omap_readl(OMAP850_IO_CONF_9) & ~0x00000001, OMAP850_IO_CONF_9);

	/* configure spi setup registers */
	omap_writew(0xfffe, OMAP850_SPI2_BASE + 0x02);
	omap_writew(0x0000, OMAP850_SPI2_BASE + 0x08);
	omap_writew(0x7ff8, OMAP850_SPI2_BASE + 0x0e);
}

static void __init htcprophet_init(void)
{
  printk("HTC Prophet init.\n");
#ifdef CONFIG_EFB_DEBUG
  efb_putstr("HTC init");
#endif
  omap_board_config = htcprophet_config;
  omap_board_config_size = ARRAY_SIZE(htcprophet_config);
  platform_add_devices(devices, ARRAY_SIZE(devices));

  htcprophet_disable_watchdog();

  htcprophet_usb_enable();
  htcprophet_spi_mux();
  htcprophet_i2c_init();
#ifdef CONFIG_HTC_MMC
  htc_mmc_init();
#endif

  // Add UART2 clock activation
  omap_writel(omap_readl(0xfffe0834) | 0x800, 0xfffe0834);
  // Add UART1 clock activation
  omap_writel(omap_readl(0xfffe0834) | 0x200, 0xfffe0834);
  omap_serial_init();

  /* For testing.. Disable for now */
  spi_register_board_info(htcprophet_spi_board_info,
			ARRAY_SIZE(htcprophet_spi_board_info));
  /* omap_free_gpio(76); */
}

static void __init htcprophet_init_irq(void)
{
	printk("htcprophet_init_irq.\n");
	omap1_init_common_hw();
	omap_init_irq();
	omap_gpio_init();
}

MACHINE_START(HTCPROPHET, "HTC Prophet")
        /* Maintainer: Unai Uribarri <unaiur@gmail.com> */
		  /* Maintainer: linwizard.sourceforge.net */
        .phys_io        = 0xfff00000,
        .io_pg_offst    = ((0xfef00000) >> 18) & 0xfffc,
        .boot_params    = 0x10000100,
        .map_io         = htcprophet_map_io,
        .init_irq       = htcprophet_init_irq,
        .init_machine   = htcprophet_init,
        .timer          = &omap_timer,
MACHINE_END
