/*
 * TSC2003 Touch Screen Driver
 * Modified from the Kaiser driver by Martin Johnson
 * Based on TSC2003 driver by Bill Gatliff
 */

/*
 * This driver accepts screen resolution
 * and pointercal data (from tslib)
 * to do its calibration through a /sys interface,
 * based on the TSC2046 driver.
 *
 * This is only a temporary solution untill
 * a better solution is available.
 */

/*
 * Having a TSC2003 driver is perhaps redundant
 * the driver for TSC2007 is already available
 * and maintained in the mainline kernel,
 * and it can be made to work with this chipset.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/string.h>
#include <linux/bcd.h>
#include <linux/list.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/input.h>
#include <linux/platform_device.h>
#include <asm/delay.h>
#include <linux/delay.h>


#include <asm/arch/gpio.h>

#define DRIVER_NAME "tsc2003-ts"

enum ts_pd {
	PD_POWERDOWN = 0, /* penirq */
	PD_IREFOFF_ADCON = 1, /* no penirq */
	PD_IREFON_ADCOFF = 2, /* penirq */
	PD_IREFON_ADCON = 3, /* no penirq */
	PD_PENIRQ_ARM = PD_IREFON_ADCOFF,
	PD_PENIRQ_DISARM = PD_IREFON_ADCON,
};

enum ts_m {
	M_12BIT = 0,
	M_8BIT = 1
};

enum ts_cmd {
	MEAS_TEMP0 = 0,
	MEAS_VBAT1 = 1,
	MEAS_IN1 = 2,
	MEAS_TEMP1 = 4,
	MEAS_VBAT2 = 5,
	MEAS_IN2 = 6,
	ACTIVATE_NX_DRIVERS = 8,
	ACTIVATE_NY_DRIVERS = 9,
	ACTIVATE_YNX_DRIVERS = 10,
	MEAS_XPOS = 12,
	MEAS_YPOS = 13,
	MEAS_Z1POS = 14,
	MEAS_Z2POS = 15
};

#define TS_CMD(cn,pdn,m) (((cn) << 4) | ((pdn) << 2) | ((m) << 1))

#define ADC_MAX ((1 << 12) - 1)

#define MAX_X 240
#define MAX_Y 320

// ADC MIN/MAX Values
int df_MIN_X=288;
int df_MAX_X=3652;

int df_MIN_Y1=320; 
int df_MAX_Y1=3565;

struct ts_data {
	struct i2c_client *client;
	struct input_dev *idev;
	struct timer_list penirq_timer;
	struct semaphore sem;
	struct task_struct *tstask;
	struct completion tstask_completion;
	enum ts_pd pd;
	enum ts_m m;
	int vbat1;
	int vbat2;
	int temp0;
	int temp1;
	int in1;
	int in2;
	int x;
	int y;
	spinlock_t lock;
	u16 use_pointercal:1;
	int pointercal[7];
	u16 x_res;
	u16 y_res;
};

#define PA d->pointercal[0]
#define PB d->pointercal[1]
#define PC d->pointercal[2]
#define PD d->pointercal[3]
#define PE d->pointercal[4]
#define PF d->pointercal[5]
#define PS d->pointercal[6]

static inline int ts_command (struct ts_data *data,
	enum ts_cmd cmd,
	enum ts_pd pd)
{
	char c;
	int ret;
	down(&data->sem);
	c = TS_CMD(cmd, pd, data->m);
	ret = i2c_master_send(data->client, &c, 1);
	up(&data->sem);
	return ret;
}

static int ts_read (struct ts_data *data,
	enum ts_cmd cmd,
	enum ts_pd pd,
	int *val)
{
	char c;
	char d[2];
	int ret;

	c = TS_CMD(cmd, pd, data->m);
	ret = i2c_master_send(data->client, &c, 1);

	udelay(20);
	ret = i2c_master_recv(data->client, d, data->m == M_12BIT ? 2 : 1);

	if (val)
	{
		*val = d[0];
		*val <<= 4;
		if (data->m == M_12BIT)
			*val += (d[1] >> 4);
	}

#if defined(CONFIG_I2C_DEBUG_CHIP)
	printk(KERN_ALERT "%s: val[%x] = %d\n",
	       __FUNCTION__, cmd, (((int)d[0]) << 8) + d[1]);
#endif
	return 0;
	if (!ret) ret = -ENODEV;
	return ret;
}

static inline int ts_read_temp0 (struct ts_data *d, enum
				      ts_pd pd, int *t)
{
	return ts_read(d, MEAS_TEMP0, pd, t);
}

static inline int ts_read_temp1 (struct ts_data *d, enum
				      ts_pd pd, int *t)
{
	return ts_read(d, MEAS_TEMP1, pd, t);
}

static inline int ts_read_xpos (struct ts_data *d, enum
				     ts_pd pd, int *x)
{
	ts_command(d,ACTIVATE_NX_DRIVERS,pd);
	udelay(100);
	return ts_read(d, MEAS_XPOS, pd, x);
}

static inline int ts_read_ypos (struct ts_data *d, enum
				     ts_pd pd, int *y)
{
	ts_command(d,ACTIVATE_NY_DRIVERS,pd);
	udelay(100);
	return ts_read(d, MEAS_YPOS, pd, y);
}

static inline int ts_read_pressure (struct ts_data *d, enum
					 ts_pd pd, int *p)
{
	ts_command(d,ACTIVATE_YNX_DRIVERS,pd);
	udelay(100);
	return ts_read(d, MEAS_Z1POS, pd, p);
}

static inline int ts_read_p2 (struct ts_data *d, enum ts_pd pd, int *p)
{
	return ts_read(d, MEAS_Z2POS, pd, p);
}
static inline int ts_read_in1 (struct ts_data *d, enum
				    ts_pd pd, int *t)
{
	return ts_read(d, MEAS_IN1, pd, t);
}

static inline int ts_read_in2 (struct ts_data *d, enum
				    ts_pd pd, int *t)
{
	return ts_read(d, MEAS_IN2, pd, t);
}

static inline int ts_read_vbat1 (struct ts_data *d, enum
				      ts_pd pd, int *t)
{
	return ts_read(d, MEAS_VBAT1, pd, t);
}

static inline int ts_read_vbat2 (struct ts_data *d, enum
				      ts_pd pd, int *t)
{
	return ts_read(d, MEAS_VBAT2, pd, t);
}

static inline int ts_powerdown (struct ts_data *d)
{
	/* we don't have a distinct powerdown command,
	   so do a benign read with the PD bits cleared */
	//char c=0xB2;
	//return i2c_master_send(d->client, &c, 1);
	return ts_read(d, 11, PD_POWERDOWN, 0);
}

void tsc2003_ts_init_client (struct i2c_client *client)
{
	struct ts_data *data = i2c_get_clientdata(client);
	
	data->pd = PD_PENIRQ_DISARM;
	data->m = M_12BIT;
	return;
}

static ssize_t tsc2003_ts_use_pointercal_show(struct device *dev,
				       struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);
	return sprintf(buf, "%u\n", d->use_pointercal);
}

static ssize_t tsc2003_ts_use_pointercal_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);
	char *endp;
	int i;

	i = simple_strtoul(buf, &endp, 10);
	spin_lock_irq(&d->lock);

   if (i)
      d->use_pointercal = 1;
	else
      d->use_pointercal = 0;

	spin_unlock_irq(&d->lock);

	return count;
}

static DEVICE_ATTR(use_pointercal, 0644, tsc2003_ts_use_pointercal_show, tsc2003_ts_use_pointercal_store);

static ssize_t tsc2003_ts_pointercal_show(struct device *dev,
                   struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);

	return sprintf(buf, "%d %d %d %d %d %d %d\n", PA, PB, PC, PD, PE, PF, PS);
}

static ssize_t tsc2003_ts_pointercal_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);
	char *endp;
	int i;
   unsigned int index = 0;

	spin_lock_irq(&d->lock);

   endp = buf;
   for (index = 0; index < 7; index++) {
      /* Grab the numbers */
      i = simple_strtol(endp, &endp, 10);
      d->pointercal[index] = i;

      /* Skip the space */
      endp++;
   }

	spin_unlock_irq(&d->lock);

	return count;
}

static DEVICE_ATTR(pointercal, 0644, tsc2003_ts_pointercal_show, tsc2003_ts_pointercal_store);

static ssize_t tsc2003_ts_screen_res_show(struct device *dev,
                   struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);

	return sprintf(buf, "%u %u\n", d->x_res, d->y_res);
}

static ssize_t tsc2003_ts_screen_res_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ts_data *d = i2c_get_clientdata(client);

	char *endp;
	unsigned int i;

	spin_lock_irq(&d->lock);

   endp = buf;

   /* Grab the x res */
   i = simple_strtoul(endp, &endp, 10);
   d->x_res = i;

   /* Skip the space */
   endp++;

   /* Grab the y res */
   i = simple_strtoul(endp, &endp, 10);
   d->y_res = i;

   /* Update the device parameters */
	input_set_abs_params(d->idev, ABS_X, 0, d->x_res - 1, 0, 0);
	input_set_abs_params(d->idev, ABS_Y, 0, d->y_res - 1, 0, 0);

	spin_unlock_irq(&d->lock);

	return count;
}

static DEVICE_ATTR(screen_res, 0644, tsc2003_ts_screen_res_show, tsc2003_ts_screen_res_store);


static int tsts_thread (void *v)
{
	struct ts_data *d = v;
	struct task_struct *tsk = current;
	int pendown=0;

	d->tstask = tsk;

	daemonize(DRIVER_NAME "tsd");
	allow_signal(SIGUSR1);

	complete(&d->tstask_completion);

	printk(KERN_INFO "%s: address 0x%x\n",
	       __FUNCTION__, d->client->addr);

	while (!signal_pending(tsk))
	{
		unsigned int x, y, p, p2;
		int xc,yc;
		int xN, yN;
		ts_read_xpos(d, PD_PENIRQ_DISARM, &x);
		ts_read_ypos(d, PD_PENIRQ_DISARM, &y);
		ts_read_pressure(d, PD_PENIRQ_DISARM, &p);
		ts_read_p2(d, PD_PENIRQ_DISARM, &p2);
		ts_read_vbat1(d, PD_PENIRQ_DISARM, 0); // avoid irq glitch
		

		if (p > 0x100 ){
		spin_lock_irq(&d->lock);
			if (d->use_pointercal)
			{
				/* Transform pen state for Android */
				xN = ((PA * x) + (PB * y) + PC) / PS;
				yN = ((PD * x) + (PE * y) + PF) / PS;
				x = xN;
				y = yN;
			}
		spin_unlock_irq(&d->lock);
			xc = 240 -((x - df_MIN_X) * MAX_X)/(df_MAX_X-df_MIN_X); //X goes 0-240
			yc = ((y - df_MIN_Y1) * MAX_Y)/(df_MAX_Y1-df_MIN_Y1); // Y goes 0-320

			if (yc < MAX_Y && (d->x != xc || d->y != yc))
			{
				d->x = xc;
				d->y = yc;
#ifdef CONFIG_INPUT_EVBUG				
				printk(KERN_ALERT "%d %d\n", xc, yc);
#endif				
				pendown = 1;
				input_report_abs(d->idev, ABS_X, xc);
				input_report_abs(d->idev, ABS_Y, yc);
				input_report_abs(d->idev, ABS_PRESSURE, p);
				input_report_key(d->idev, BTN_TOUCH, 1);
				input_sync(d->idev);
			}

		} else if (pendown == 1) {
			d->x = 0;
			d->y = 0;
			input_report_key(d->idev, BTN_TOUCH, 0);
			input_report_abs(d->idev, ABS_PRESSURE, 0);
			input_sync(d->idev);
			pendown = 0;
		}
#ifdef CONFIG_INPUT_EVBUG
		schedule_timeout_interruptible(msecs_to_jiffies(100));
#else
		schedule_timeout_interruptible(msecs_to_jiffies(30));
#endif		
	}

	d->tstask = NULL;
	complete_and_exit(&d->tstask_completion, 0);
}


static int ts_idev_open (struct input_dev *idev)
{
	struct ts_data *d = idev->private;
	int ret = 0;

	if (down_interruptible(&d->sem))
		return -EINTR;

	if (d->tstask)
		panic(DRIVER_NAME "tsd already running (!). abort.");

	init_completion(&d->tstask_completion);

	ret = kernel_thread(tsts_thread, d, CLONE_KERNEL);
	if (ret >= 0) {
		wait_for_completion(&d->tstask_completion);
		ret = 0;
	}

	up(&d->sem);
	return ret;
}

static void ts_idev_close (struct input_dev *idev)
{
	struct ts_data *d = idev->private;
	down_interruptible(&d->sem);
	if (d->tstask)
	{
		send_sig(SIGUSR1, d->tstask, 1);
		wait_for_completion(&d->tstask_completion);
	}
	up(&d->sem);
	return;
}

static int ts_driver_register (struct ts_data *data)
{
	int ret = 0;

	init_MUTEX(&data->sem);

	data->idev = input_allocate_device();
	if(!data->idev)
	{
		return -ENOMEM;
	}
	data->idev->private = data;
	data->idev->name = DRIVER_NAME;
	set_bit(EV_SYN, data->idev->evbit);
	set_bit(EV_KEY, data->idev->evbit);
	set_bit(BTN_TOUCH, data->idev->keybit);
	set_bit(EV_ABS, data->idev->evbit);
	
	data->idev->open = ts_idev_open;
	data->idev->close = ts_idev_close;
	data->idev->absbit[ABS_X] = BIT(ABS_X);
	data->idev->absbit[ABS_Y] = BIT(ABS_Y);
	data->idev->absbit[ABS_PRESSURE] = BIT(ABS_PRESSURE);
	input_set_abs_params(data->idev, ABS_X, 0, MAX_X, 0, 0);
	input_set_abs_params(data->idev, ABS_Y, 0, MAX_Y, 0, 0);
    	input_set_abs_params(data->idev, ABS_PRESSURE, 0, 1, 0, 0);

	ret = input_register_device(data->idev);
	if(ret)
	{
		input_free_device(data->idev);
		return ret;
	}
	return ret;
}

static int  tsc2003_ts_probe(struct i2c_client *client)
{
	int ret;
	struct ts_data *data;

	printk(KERN_INFO "tsc2003 i2c touch screen\n");
	
	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		printk(KERN_ERR "tsc2003_ts_probe: need I2C_FUNC_I2C\n");
		ret = -ENODEV;
		goto err;
	}
	data = kcalloc(1, sizeof(*data), GFP_KERNEL);
	if (!data) {
		ret = -ENOMEM;
		goto err;
	}
	
	data->client = client;
	i2c_set_clientdata(client, data);

	tsc2003_ts_init_client(client);

	ts_powerdown(data);

	ret = ts_driver_register(data);
	if(ret < 0) {
		printk(KERN_ALERT "driver_register failed\n");
		goto err;
	}

	if (device_create_file(&client->dev, &dev_attr_use_pointercal) < 0)
		goto err1;
	if (device_create_file(&client->dev, &dev_attr_pointercal) < 0)
		goto err2;
	if (device_create_file(&client->dev, &dev_attr_screen_res) < 0)
		goto err3;
	
	return ret;

err3:
   device_remove_file(&client->dev, &dev_attr_screen_res);
err2:
   device_remove_file(&client->dev, &dev_attr_pointercal);
err1:
   device_remove_file(&client->dev, &dev_attr_use_pointercal);
err:
	kfree(client);
	return ret;
}

#define tsc2003_ts_suspend NULL
#define tsc2003_ts_resume NULL


static int __devexit tsc2003_ts_remove(struct i2c_client *client)
{
	printk(KERN_ALERT  "tsc2003_ts_remove\n");
	struct ts_data *ts = i2c_get_clientdata(client);
	input_unregister_device(ts->idev);
	kfree(ts);
	return 0;
}

static struct i2c_driver tsc2003_ts_driver = {
	.probe		= tsc2003_ts_probe,
	.remove		= tsc2003_ts_remove,
	.suspend	= tsc2003_ts_suspend,
	.resume		= tsc2003_ts_resume,
	.driver = {
		.name	= DRIVER_NAME,
	},
};

static int __devinit tsc2003_ts_init(void)
{
	return i2c_add_driver(&tsc2003_ts_driver);
}

static void __exit tsc2003_ts_exit(void)
{
	i2c_del_driver(&tsc2003_ts_driver);
}

MODULE_AUTHOR("Amer Khaznadar <musikmonk@gmail.com>");
MODULE_DESCRIPTION("TSC2003 Touch Screen");
MODULE_LICENSE("GPL");

module_init(tsc2003_ts_init);
module_exit(tsc2003_ts_exit);
