/*
 * Fake battery driver -- based on the DS2760 chip driver
 *
 * Copyright © 2009 Robert Jarzmik
 *
 * Based on work Copyright © 2007 Anton Vorontsov
 *        2004-2007 Matt Reimer
 *        2004 Szabolcs Gyurko
 *
 * Use consistent with the GNU GPL is permitted,
 * provided that this copyright notice is
 * preserved in its entirety in all copies and derived works.
 *
 * Author:  Robert Jarzmik
 *     August 2009
 *
 *     Anton Vorontsov <cbou@mail.ru>
 *     February 2007
 *
 *     Matt Reimer <mreimer@vpop.net>
 *     April 2004, 2005, 2007
 *
 *     Szabolcs Gyurko <szabolcs.gyurko@tlt.hu>
 *     September 2004
 */

#include <linux/module.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/workqueue.h>
#include <linux/pm.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>


struct fake_device_info {
   struct device *dev;

   /* DS2760 data, valid after calling ds2760_battery_read_status() */
   unsigned long update_time; /* jiffies when data read */
   int voltage_raw;     /* units of 4.88 mV */
   int voltage_uV;         /* units of µV */
   int current_raw;     /* units of 0.625 mA */
   int current_uA;         /* units of µA */
   int accum_current_raw;     /* units of 0.25 mAh */
   int accum_current_uAh;     /* units of µAh */
   int temp_raw;        /* units of 0.125 °C */
   int temp_C;       /* units of 0.1 °C */
   int rated_capacity;     /* units of µAh */
   int rem_capacity;    /* percentage */
   int full_active_uAh;    /* units of µAh */
   int empty_uAh;       /* units of µAh */
   int life_sec;        /* units of seconds */
   int charge_status;      /* POWER_SUPPLY_STATUS_* */

   int full_counter;
   struct power_supply bat;
   struct device *fake_dev;
};


/* Some batteries have their rated capacity stored a N * 10 mAh, while
 * others use an index into this table. */
static int rated_capacities[] = {
   0,
   920,  /* Samsung */
   920,  /* BYD */
   920,  /* Lishen */
   920,  /* NEC */
   1440, /* Samsung */
   1440, /* BYD */
   1440, /* Lishen */
   1440, /* NEC */
   2880, /* Samsung */
   2880, /* BYD */
   2880, /* Lishen */
   2880  /* NEC */
};

static int fake_battery_get_property(struct power_supply *psy,
                   enum power_supply_property psp,
                   union power_supply_propval *val)
{
   switch (psp) {
   case POWER_SUPPLY_PROP_STATUS:
      val->intval = POWER_SUPPLY_STATUS_NOT_CHARGING;
      return 0;
   default:
      break;
   }

   switch (psp) {
   case POWER_SUPPLY_PROP_VOLTAGE_NOW:
      val->intval = 3700;
      break;
   case POWER_SUPPLY_PROP_CURRENT_NOW:
      val->intval = 89000000;
      break;
   case POWER_SUPPLY_PROP_CHARGE_FULL_DESIGN:
      val->intval = 1500;
      break;
   case POWER_SUPPLY_PROP_CHARGE_FULL:
      val->intval = 1500;
      break;
   case POWER_SUPPLY_PROP_CHARGE_EMPTY:
      val->intval = 250;
      break;
   case POWER_SUPPLY_PROP_CHARGE_NOW:
      val->intval = 1250;
      break;
   case POWER_SUPPLY_PROP_TEMP:
      val->intval = 260;
      break;
   case POWER_SUPPLY_PROP_HEALTH:
      val->intval = POWER_SUPPLY_HEALTH_GOOD;
      break;
   case POWER_SUPPLY_PROP_PRESENT:
      val->intval = 1;
      break;
   case POWER_SUPPLY_PROP_ONLINE:
      val->intval = 1;
      break;
   case POWER_SUPPLY_PROP_CAPACITY:
      val->intval = 100;
      break;
   case POWER_SUPPLY_PROP_TECHNOLOGY:
      val->intval = POWER_SUPPLY_TECHNOLOGY_LION;
      break;
   default:
      return -EINVAL;
   }

   return 0;
}

static enum power_supply_property fake_battery_props[] = {
   POWER_SUPPLY_PROP_STATUS,
        POWER_SUPPLY_PROP_HEALTH,
        POWER_SUPPLY_PROP_PRESENT,
        POWER_SUPPLY_PROP_CAPACITY,
   POWER_SUPPLY_PROP_VOLTAGE_NOW,
   POWER_SUPPLY_PROP_CURRENT_NOW,
   POWER_SUPPLY_PROP_CHARGE_FULL_DESIGN,
   POWER_SUPPLY_PROP_CHARGE_FULL,
   POWER_SUPPLY_PROP_CHARGE_EMPTY,
   POWER_SUPPLY_PROP_CHARGE_NOW,
   POWER_SUPPLY_PROP_TECHNOLOGY,
   POWER_SUPPLY_PROP_TEMP,
};

static int fake_battery_probe(struct platform_device *pdev)
{
   int retval = 0;
   struct fake_device_info *di;
        printk("Probing Fake battery driver\n");

   di = kzalloc(sizeof(*di), GFP_KERNEL);
   if (!di) {
      retval = -ENOMEM;
      goto di_alloc_failed;
   }

   platform_set_drvdata(pdev, di);

   di->dev     = &pdev->dev;
   di->fake_dev        = pdev->dev.parent;
   di->bat.name      = pdev->dev.bus_id;
   di->bat.type      = POWER_SUPPLY_TYPE_BATTERY;
   di->bat.properties     = fake_battery_props;
   di->bat.num_properties = ARRAY_SIZE(fake_battery_props);
   di->bat.get_property   = fake_battery_get_property;

   di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;

   retval = power_supply_register(&pdev->dev, &di->bat);
   if (retval) {
      dev_err(di->dev, "failed to register battery\n");
      goto batt_failed;
   }
        printk("Registered Fake battery driver\n");

   goto success;

workqueue_failed:
   power_supply_unregister(&di->bat);
batt_failed:
   kfree(di);
di_alloc_failed:
success:
   return retval;
}

static int fake_battery_remove(struct platform_device *pdev)
{
   struct fake_device_info *di = platform_get_drvdata(pdev);

   power_supply_unregister(&di->bat);

   return 0;
}

#ifdef CONFIG_PM

static int fake_battery_suspend(struct platform_device *pdev,
              pm_message_t state)
{
   struct fake_device_info *di = platform_get_drvdata(pdev);

   di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;

   return 0;
}

static int fake_battery_resume(struct platform_device *pdev)
{
   struct fake_device_info *di = platform_get_drvdata(pdev);

   di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;
   power_supply_changed(&di->bat);

   return 0;
}

#else

#define fake_battery_suspend NULL
#define fake_battery_resume NULL

#endif /* CONFIG_PM */

static struct platform_device *bat_pdev;

static struct power_supply fake_bat = {
   .name = "battery",
   .properties = fake_battery_props,
   .num_properties = ARRAY_SIZE(fake_battery_props),
   .get_property = fake_battery_get_property,
   .use_for_apm = 1,
};

static int __init fake_battery_init(void)
{
   /* Ignore the status. It doesn't actually matter */
        int ret;
   bat_pdev = platform_device_register_simple("fake-battery", 0, NULL, 0);

   ret = power_supply_register(&bat_pdev->dev, &fake_bat);
   if (ret)
      goto battery_failed;

   goto success;

battery_failed:
   platform_device_unregister(bat_pdev);
success:
   return ret;
}

static void __exit fake_battery_exit(void)
{
   power_supply_unregister(&fake_bat);
   platform_device_unregister(bat_pdev);

}

module_init(fake_battery_init);
module_exit(fake_battery_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Robert Jarzmik");
MODULE_DESCRIPTION("Fake battery driver");
