/*
 *  Backlight Driver for HTC Wizard and HTC Herald using the I2C HTCCPLD.
 *
 *  Copyright (c) 2008-2009 Angelo Arrifano <miknix@gmail.com>
 *  Copyright (C) 2009 Cory Maccarrone <darkstar6262@gmail.com>
 *
 *  Based on drivers/video/backlight/hp680_bl.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/backlight.h>
#include <linux/fb.h>
#include <linux/i2c/i2c-htcpld.h>

static int htcwiz_bl_set_brightness(struct backlight_device *bl_dev)
{
   u8 dim_val = 0;
   u8 panel_power = 1;
   u8 chip_val = 0;

	if (bl_dev->props.power != FB_BLANK_UNBLANK ||
			bl_dev->props.fb_blank != FB_BLANK_UNBLANK) {
      panel_power = 0;
   } else {
      panel_power = 1;
   }

   switch (bl_dev->props.brightness) {
      case 0:
         /* Maximum dim */
         dim_val = 0 << 1;
         break;

      case 1:
         /* Dim level 1 */
         dim_val = 2 << 1;
         break;

      case 2:
         /* Dim level 2 */
         dim_val = 1 << 1;
         break;

      case 3:
         /* Max brightness */
         dim_val = 3 << 1;
         break;
   }

   chip_val = htcpld_chip_get(0x04);

   /* Clear the current bits */
   chip_val &= 0xf1;

   /* Set the new value */
   chip_val |= dim_val;

   /* Set the panel power */
   chip_val |= (panel_power << 3);

   htcpld_chip_set(0x04, chip_val);
	return 0;
}

static int htcwiz_bl_get_brightness(struct backlight_device *bl_dev)
{
   u8 dim_val_raw = 0;
   u8 dim_val = 0;

   dim_val_raw = htcpld_chip_get(0x04);
   dim_val_raw &= 0x6;
   dim_val_raw = dim_val_raw >> 1;

   switch (dim_val_raw) {
      case 3:
         /* Max brightness */
         dim_val = 3;
         break;

      case 2:
         /* Level 1 */
         dim_val = 1;
         break;

      case 1:
         /* Level 2 */
         dim_val = 2;
         break;

      case 0:
         /* Level 0 */
         dim_val = 0;
         break;
   }

	return dim_val;
}

static struct backlight_ops htcwiz_bl_ops = {
	.get_brightness = htcwiz_bl_get_brightness,
	.update_status  = htcwiz_bl_set_brightness,
};

static int __init htcwiz_bl_probe(struct platform_device *pdev)
{
	struct backlight_device *bl_dev;

	bl_dev = backlight_device_register ("htcwiz_bl", &pdev->dev, NULL,
		    &htcwiz_bl_ops);
	/* FIXME:
	if (IS_ERR(bl_dev))
		return PTR_ERR(bl_dev);*/

	platform_set_drvdata(pdev, bl_dev);

	bl_dev->props.max_brightness = HTCI2CCPLD_BL_LMAX;
	bl_dev->props.brightness = htcwiz_bl_get_brightness(bl_dev);

	return 0;
}

static int htcwiz_bl_remove(struct platform_device *pdev)
{
	struct backlight_device *bl_dev = platform_get_drvdata(pdev);

	backlight_device_unregister(bl_dev);

	return 0;
}

static struct platform_driver htcwiz_bl_driver = {
	.probe		= htcwiz_bl_probe,
	.remove		= htcwiz_bl_remove,
	.driver		= {
		.name	= "htcwiz_bl",
	},
};

static struct platform_device *htcwiz_bl_device;

static int __init htcwiz_bl_init(void)
{
	int ret;

	ret = platform_driver_register(&htcwiz_bl_driver);
	if (!ret) {
		htcwiz_bl_device = platform_device_alloc("htcwiz_bl", -1);
		if (!htcwiz_bl_device)
			return -ENOMEM;

		ret = platform_device_add(htcwiz_bl_device);

		if (ret) {
			platform_device_put(htcwiz_bl_device);
			platform_driver_unregister(&htcwiz_bl_driver);
		}
	}
	printk("htcwiz_bl: HTC Wizard backlight driver\n");

	return ret;
}

static void __exit htcwiz_bl_exit(void)
{
	platform_device_unregister(htcwiz_bl_device);
 	platform_driver_unregister(&htcwiz_bl_driver);
}

module_init(htcwiz_bl_init);
module_exit(htcwiz_bl_exit);

MODULE_AUTHOR("Angelo Arrifano <miknix@gmail.com>");
MODULE_DESCRIPTION("HTC Wizard Backlight driver");
MODULE_LICENSE("GPL");
