/*
 * File: drivers/video/omap/lcd-htcgene.c
 *
 * LCD panel support for the HTC Gene
 *
 * Copyright (C) wing-linux.sourceforge.net
 * Author: Cory Maccarrone <darkstar6262@gmail.com>
 * Based on lcd-htcwizard.c by Angelo Arrifano <miknix@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/platform_device.h>

#include <asm/arch/gpio.h>
#include <asm/arch/omapfb.h>

#define HTCGENE_GPIO_PANEL_ID0 44 /* LCD Panel identification */
#define HTCGENE_GPIO_PANEL_ID1 55 /* LCD Panel identification */

static uint def_screen = 99;
static uint def_custom_screen = 0;
static uint def_xres = 0;
static uint def_hsw = 0;
static uint def_hfp = 0;
static uint def_hbp = 0;
static uint def_yres = 0;
static uint def_vsw = 0;
static uint def_vfp = 0;
static uint def_vbp = 0;
static uint def_ivs = 0;
static uint def_ihs = 0;
static uint def_ipc = 0;
static uint def_pixelclock = 0;
static uint def_disable_at_init = 1;

#define       OMAP_LCDC_CONTROL               (0xfffec000 + 0x00)
#define       OMAP_LCDC_FB_X                  (0xfffec000 + 0x04)
#define       OMAP_LCDC_FB_Y                  (0xfffec000 + 0x08)
#define       OMAP_LCDC_FB_FLAGS              (0xfffec000 + 0x0c)
#define       OMAP_LCDC_CONTROL               (0xfffec000 + 0x00)
#define       OMAP_LCDC_CONTROL               (0xfffec000 + 0x00)
#define       OMAP_LCDC_STATUS                (0xfffec000 + 0x10)
#define       OMAP_DMA_LCD_CCR                (0xfffee300 + 0xc2)
#define       OMAP_DMA_LCD_CTRL               (0xfffee300 + 0xc4)
#define       OMAP_LCDC_CTRL_LCD_EN           (1 << 0)
#define       OMAP_LCDC_STAT_DONE             (1 << 0)

static int htcgene_panel_init(struct lcd_panel *panel,
                                struct omapfb_device *fbdev)
{
   u32 reg;

   if (def_disable_at_init) {
      printk("About to disable FB device\n");

      /* disable controller if active */
      reg = omap_readl(OMAP_LCDC_CONTROL);
      printk("(before disable) OMAP_LCDC_CONTROL = 0x%x\n", reg);

      if (reg & OMAP_LCDC_CTRL_LCD_EN) {
         reg &= ~OMAP_LCDC_CTRL_LCD_EN;
         omap_writel(reg, OMAP_LCDC_CONTROL);

         /* wait for end of frame */
         while (!(omap_readl(OMAP_LCDC_STATUS) & OMAP_LCDC_STAT_DONE));

         /* turn off DMA */
         reg = omap_readw(OMAP_DMA_LCD_CCR);
         reg &= ~(1 << 7);
         omap_writew(reg, OMAP_DMA_LCD_CCR);

         reg = omap_readw(OMAP_DMA_LCD_CTRL);
         reg &= ~(1 << 8);
         omap_writew(reg, OMAP_DMA_LCD_CTRL);
      }
   } else {
      printk("Disable at init disabled\n");
   }

   reg = omap_readl(OMAP_LCDC_CONTROL);
   printk("(after disable) OMAP_LCDC_CONTROL = 0x%x\n", reg);

	return 0;
}

static void htcgene_panel_cleanup(struct lcd_panel *panel)
{
}

static int htcgene_panel_enable(struct lcd_panel *panel)
{
   u32 reg;

   printk("omapfb: about to enable framebuffer\n");

   /* enable controller */
   reg = omap_readl(OMAP_LCDC_CONTROL);
   printk("(before enable) OMAP_LCDC_CONTROL = 0x%x\n", reg);

   if (!(reg & OMAP_LCDC_CTRL_LCD_EN)) {
      /* wait for end of frame */
      while (!(omap_readl(OMAP_LCDC_STATUS) & OMAP_LCDC_STAT_DONE));

      /* turn on DMA */
      reg = omap_readw(OMAP_DMA_LCD_CCR);
      reg |= (1 << 7);
      omap_writew(reg, OMAP_DMA_LCD_CCR);

      reg = omap_readw(OMAP_DMA_LCD_CTRL);
      reg |= (1 << 8);
      omap_writew(reg, OMAP_DMA_LCD_CTRL);

      /* enable the panel */
      reg |= OMAP_LCDC_CTRL_LCD_EN;
      omap_writel(reg, OMAP_LCDC_CONTROL);
   }

   reg = omap_readl(OMAP_LCDC_CONTROL);
   printk("(after enable) OMAP_LCDC_CONTROL = 0x%x\n", reg);

	return 0;
}

static void htcgene_panel_disable(struct lcd_panel *panel)
{
   /* LCD register definition (maybe to move somewhere else */
   u32 reg;
   /* disable controller */
   reg = omap_readl(OMAP_LCDC_CONTROL);
   if (reg & OMAP_LCDC_CTRL_LCD_EN) {
      reg &= ~OMAP_LCDC_CTRL_LCD_EN;
      omap_writel(reg, OMAP_LCDC_CONTROL);

      /* wait for end of frame */
      while (!(omap_readl(OMAP_LCDC_STATUS) & OMAP_LCDC_STAT_DONE));
   }
}

static unsigned long htcgene_panel_get_caps(struct lcd_panel *panel)
{
	return 0;
}

/* Found on WIZ200 (miknix) and some HERA110 models (darkstar62) */
struct lcd_panel htcgene_panel_1 = {
	.name		= "lcd_gene",
	.config		= OMAP_LCDC_PANEL_TFT |
	              OMAP_LCDC_INV_HSYNC |
                 OMAP_LCDC_INV_VSYNC |
					  OMAP_LCDC_INV_PIX_CLOCK,
	.bpp		= 16,
	.data_lines	= 16,
	.x_res		= 240,
	.y_res		= 320,
	.pixel_clock = 6093,
   .pcd     = 0, /* 15 */
	.hsw		= 10,
	.hfp		= 10,
	.hbp		= 20,
	.vsw		= 3,
	.vfp		= 2,
	.vbp		= 2,

	.init		= htcgene_panel_init,
	.cleanup	= htcgene_panel_cleanup,
	.enable		= htcgene_panel_enable,
	.disable	= htcgene_panel_disable,
	.get_caps	= htcgene_panel_get_caps,
};

/* Found on some HERA110 models (darkstar62) */
struct lcd_panel htcgene_panel_2 = {
	.name		= "lcd_gene2",
	.config		= OMAP_LCDC_PANEL_TFT |
	              OMAP_LCDC_INV_HSYNC |
                 OMAP_LCDC_INV_VSYNC,
	.bpp		= 16,
	.data_lines	= 16,
	.x_res		= 240,
	.y_res		= 320,
	.pixel_clock = 5712,
   .pcd     = 0, /* 16 */
	.hsw		= 4,
	.hfp		= 2,
	.hbp		= 10,
	.vsw		= 2,
	.vfp		= 57,
	.vbp		= 5,

	.init		= htcgene_panel_init,
	.cleanup	= htcgene_panel_cleanup,
	.enable		= htcgene_panel_enable,
	.disable	= htcgene_panel_disable,
	.get_caps	= htcgene_panel_get_caps,
};

/* Found on an official HTC Herald */
struct lcd_panel htcgene_panel_3 = {
	.name		= "lcd_gene3",
	.config		= OMAP_LCDC_PANEL_TFT |
	              OMAP_LCDC_INV_HSYNC |
                 OMAP_LCDC_INV_VSYNC,
	.bpp		= 16,
	.data_lines	= 16,
	.x_res		= 240,
	.y_res		= 320,
	.pixel_clock = 4431, /* Different base clock */
   .pcd     = 0, /* 16 */
	.hsw		= 4,
	.hfp		= 2,
	.hbp		= 9,
	.vsw		= 2,
	.vfp		= 57,
	.vbp		= 5,

	.init		= htcgene_panel_init,
	.cleanup	= htcgene_panel_cleanup,
	.enable		= htcgene_panel_enable,
	.disable	= htcgene_panel_disable,
	.get_caps	= htcgene_panel_get_caps,
};

void gene_decode_mode(struct lcd_panel *panel, u32 xreg, u32 yreg, u32 freg) {
   panel->x_res = (xreg & 0x3ff) + 1;
   panel->hsw   = ((xreg & 0x3fff) >> 10) + 1;
   panel->hfp   = ((xreg & 0x3fffff) >> 16) + 1;
   panel->hbp   = ((xreg & 0x3fffffff) >> 24) + 1;

   panel->y_res = (yreg & 0x3ff) + 1;
   panel->vsw   = ((yreg & 0x3fff) >> 10) + 1;
   panel->vfp   = ((yreg & 0x3fffff) >> 16);
   panel->vbp   = ((yreg & 0x3fffffff) >> 24);

   panel->config = OMAP_LCDC_PANEL_TFT;

   panel->pcd   = (freg & 0xff) + 1;
   panel->pixel_clock = 0;
   panel->config |= ((((freg & 0xffffff) >> 20) & 1) > 0) ? OMAP_LCDC_INV_VSYNC : 0;
   panel->config |= ((((freg & 0xffffff) >> 20) & 2) > 0) ? OMAP_LCDC_INV_HSYNC : 0;
   panel->config |= ((((freg & 0xffffff) >> 20) & 4) > 0) ? OMAP_LCDC_INV_PIX_CLOCK : 0;
   panel->config |= ((((freg & 0xffffff) >> 20) & 8) > 0) ? OMAP_LCDC_INV_OUTPUT_EN : 0;
}

static int htcgene_panel_probe(struct platform_device *pdev)
{
   struct lcd_panel *panel;

   printk("LCD: FIXME: Mode timings are hard-coded into configuration!\n");

   if (def_custom_screen) {
      panel = &htcgene_panel_1;
      panel->name = "lcd_gene_custom";
      panel->x_res = def_xres;
      panel->y_res = def_yres;
      panel->pixel_clock = def_pixelclock;
      panel->hsw = def_hsw;
      panel->hfp = def_hfp;
      panel->hbp = def_hbp;
      panel->vsw = def_vsw;
      panel->vfp = def_vfp;
      panel->vbp = def_vbp;

      panel->config = (def_ihs ? OMAP_LCDC_INV_HSYNC : 0 ) |
                      (def_ivs ? OMAP_LCDC_INV_VSYNC : 0 ) |
                      (def_ipc ? OMAP_LCDC_INV_PIX_CLOCK : 0 ) |
                      OMAP_LCDC_PANEL_TFT;
   } else {
      switch (def_screen) {
         case 2:
            printk("LCD: Panel type lcd_gene3 used\n");
            panel = &htcgene_panel_3;
            break;

         case 1:
            printk("LCD: Panel type lcd_gene2 used\n");
            panel = &htcgene_panel_2;
            break;

         case 0:
            printk("LCD: Panel type lcd_gene used\n");
            panel = &htcgene_panel_1;
            break;

         default:
         {
            u32 xreg, yreg, flagreg;

            printk("LCD: Attempting autodetection\n");
            xreg = omap_readl(OMAP_LCDC_FB_X);
            yreg = omap_readl(OMAP_LCDC_FB_Y);
            flagreg = omap_readl(OMAP_LCDC_FB_FLAGS);

            panel = &htcgene_panel_1;
            gene_decode_mode(panel, xreg, yreg, flagreg);
            break;
         }
      }
   }

   printk("LCD: Panel type used:\n");
   printk("  x_res:       %d\n", panel->x_res);
   printk("  y_res:       %d\n", panel->y_res);
   printk("  pixel_clock: %d\n", panel->pixel_clock);
   printk("  pcd:         %d\n", panel->pcd);
   printk("  hsw:         %d\n", panel->hsw);
   printk("  hfp:         %d\n", panel->hfp);
   printk("  hbp:         %d\n", panel->hbp);
   printk("  vsw:         %d\n", panel->vsw);
   printk("  vfp:         %d\n", panel->vfp);
   printk("  vbp:         %d\n", panel->vbp);
   printk("  bpp:         %d\n", panel->bpp);
   printk("  data_lines:  %d\n", panel->data_lines);
   printk("  config:      %d\n", panel->config);

   omapfb_register_panel(panel);
	return 0;
}

static int htcgene_panel_remove(struct platform_device *pdev)
{
	return 0;
}

static int htcgene_panel_suspend(struct platform_device *pdev, pm_message_t mesg)
{
	return 0;
}

static int htcgene_panel_resume(struct platform_device *pdev)
{
	return 0;
}

struct platform_driver htcgene_panel_driver = {
	.probe		= htcgene_panel_probe,
	.remove		= htcgene_panel_remove,
	.suspend	= htcgene_panel_suspend,
	.resume		= htcgene_panel_resume,
	.driver		= {
		.name	= "lcd_htcgene",
		.owner	= THIS_MODULE,
	},
};

/* Process kernel command line parameters */
static void htcgene_panel_drv_setup(char *options)
{
	char *this_opt = NULL;

	printk("htcgene_panel: options %s\n", options);

	if (!options || !*options)
		return;

	while ((this_opt = strsep(&options, ",")) != NULL) {
      if (!strncmp(this_opt, "custscreen", 10))
         def_custom_screen = 1;
      else if (!strncmp(this_opt, "screen:", 7))
			def_screen = (simple_strtoul(this_opt + 7, NULL, 0));
		else if (!strncmp(this_opt, "noinitdisable", 14))
         def_disable_at_init = 0;
      else if (!strncmp(this_opt, "xres:", 5))
         def_xres = (simple_strtoul(this_opt + 5, NULL, 0));
      else if (!strncmp(this_opt, "hsw:", 4))
         def_hsw = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "hfp:", 4))
         def_hfp = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "hbp:", 4))
         def_hbp = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "yres:", 5))
         def_yres = (simple_strtoul(this_opt + 5, NULL, 0));
      else if (!strncmp(this_opt, "vsw:", 4))
         def_vsw = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "vfp:", 4))
         def_vfp = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "vbp:", 4))
         def_vbp = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "pixelclock:", 11))
         def_pixelclock = (simple_strtoul(this_opt + 11, NULL, 0));
      else if (!strncmp(this_opt, "ihs:", 4))
         def_ihs = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "ivs:", 4))
         def_ivs = (simple_strtoul(this_opt + 4, NULL, 0));
      else if (!strncmp(this_opt, "ipc:", 4))
         def_ipc = (simple_strtoul(this_opt + 4, NULL, 0));
		else
			printk("htchearld_panel: invalid option\n");
	}
}

static int htcgene_panel_drv_init(void)
{
	return platform_driver_register(&htcgene_panel_driver);
}

static void htcgene_panel_drv_cleanup(void)
{
	platform_driver_unregister(&htcgene_panel_driver);
}

__setup("genelcd=", htcgene_panel_drv_setup);

module_init(htcgene_panel_drv_init);
module_exit(htcgene_panel_drv_cleanup);

